//---------------------------------------------------------------------------

CL_SceneManager::CL_SceneManager(CL_PixelBuffer Buffer, CL_ResourceManager* Manager) : CL_Surface(Buffer)
{
	m_Canvas = new CL_Canvas();
	m_Canvas->select_surface(*this);
	m_SceneName = "";
	m_ResourceManager = Manager;
}

//---------------------------------------------------------------------------

CL_SceneManager::~CL_SceneManager()
{
	clearLayers();
}

//---------------------------------------------------------------------------
// void clearLayers()
// Removes all layer objects from the scene manager
// This deletes all layers and their child sprites

void CL_SceneManager::clearLayers()
{
	for(unsigned int i = 0; i < m_Layers.size(); i++)
	{
		if(m_Layers[i])
			delete m_Layers[i];
	}

	m_Layers.clear();
}

//---------------------------------------------------------------------------
//void composite(CL_Rect *Rect)
//Renders all layer surfaces onto the scene manager surface
//Called each draw

void CL_SceneManager::composite(CL_Rect *Rect)
{
	std::list<CL_SceneLayer*>::const_iterator it;

	m_Canvas->get_gc()->clear(CL_Color(0, 0, 0, 0));

	//Cycle through all layers to draw
	for(unsigned int i = 0; i < m_Layers.size(); i++)
	{
		CL_SceneLayer *Layer = m_Layers[i];

		if(Layer->getVisible()) //If layer visible
		{
			//Test Layer Cull

			CL_Rect LayerRect;
			LayerRect.left = (int) (Rect->left + Layer->m_Position.x);
			LayerRect.top = (int) (Rect->top + Layer->m_Position.y);
			LayerRect.right = (int) (LayerRect.left + Layer->get_width());
			LayerRect.bottom = (int) (LayerRect.top + Layer->get_height());

			if(Rect->is_overlapped(LayerRect)) //then draw layer
			{
				Layer->draw((float) LayerRect.left, (float) LayerRect.top, m_Canvas->get_gc());
			}
		}
	}

	m_Canvas->sync_surface();
}

//---------------------------------------------------------------------------
//CL_SceneLayer* createLayer(float Width, float height, CL_PixelFormat Format, float Pitch, std::string Name)
//Creates a scene layer. This is not attached to the current scene. It's just created.
//Afterwards, use addLayer() to add.

CL_SceneLayer* CL_SceneManager::createLayer(float Width, float height, CL_PixelFormat Format, float Pitch, std::string Name)
{
	CL_PixelBuffer Buffer((int) Width, (int) height, (int) Pitch, Format, NULL);
	CL_SceneLayer *Layer = new CL_SceneLayer(Buffer, Name);

	return Layer;
}

//---------------------------------------------------------------------------
//CL_SceneObject* createSceneObject(std::string Name, CL_ResourceManager* ResourceManager, float x, float y, bool visible)
//Creates a scene object. This is not attached to a layer. It's just created.
//Afterwards, use CL_SceneLayer::attachObject to add.

CL_SceneObject* CL_SceneManager::createSceneObject(std::string Name, CL_ResourceManager* ResourceManager,
												   float x, float y, bool visible)
{
	return new CL_SceneObject(Name, ResourceManager, x, y, visible);
}

//---------------------------------------------------------------------------
void CL_SceneManager::draw(float x, float y, CL_GraphicContext* context)
{
	//Manager Rect Properties; absolute top, left, width and height on screen that will be drawn
	CL_Rect Rect;
	Rect.left = (int)x;
	Rect.top = (int)y;
	Rect.right = Rect.left + get_width();
	Rect.bottom = Rect.top + get_height();

	composite(&Rect);

	CL_Surface::draw(x, y, context);
}

//---------------------------------------------------------------------------
void CL_SceneManager::draw(const CL_Rect& dest, CL_GraphicContext* context)
{
	CL_Rect Rect;
	Rect.left = 0;
	Rect.top = 0;
	Rect.right = get_width();
	Rect.bottom = get_height();

	composite(&Rect);

	CL_Surface::draw(dest, context);
}

//---------------------------------------------------------------------------
void CL_SceneManager::draw(const CL_Rect& src, const CL_Rect& dest, CL_GraphicContext* context)
{
	CL_Rect Rect = src;
	composite(&Rect);

	CL_Surface::draw(src, dest, context);
}

//---------------------------------------------------------------------------
void CL_SceneManager::draw(const CL_Surface_DrawParams1& params1, CL_GraphicContext* context)
{
	CL_Rect Rect;
	Rect.left = (int)params1.srcX;
	Rect.top = (int)params1.srcY;
	Rect.right = Rect.left + get_width();
	Rect.bottom = Rect.top + get_height();

	composite(&Rect);

	CL_Surface::draw(params1, context);
}

//---------------------------------------------------------------------------
void CL_SceneManager::draw(const CL_Surface_DrawParams2& params2, CL_GraphicContext* context)
{
	CL_Rect Rect;
	Rect.left = (int)params2.srcX;
	Rect.top = (int)params2.srcY;
	Rect.right = Rect.left + params2.srcWidth;
	Rect.bottom = Rect.top + params2.srcHeight;

	composite(&Rect);

	CL_Surface::draw(params2, context);
}

//---------------------------------------------------------------------------
//CL_SceneLayer* findLayer(std::string Name)
//Finds first layer in the scene with matching name

CL_SceneLayer* CL_SceneManager::findLayer(std::string Name)
{
	for(unsigned int i = 0; i < m_Layers.size(); i++)
	{
		if(Name == m_Layers[i]->getName())
			return m_Layers[i];
	}

	return NULL;
}

//---------------------------------------------------------------------------
//void swapLayers(unsigned int Index1, unsigned int Index2)
//Exchanges the z-order of two layers

void CL_SceneManager::swapLayers(unsigned int Index1, unsigned int Index2)
{
	std::swap(m_Layers[Index1], m_Layers[Index2]);
}

//---------------------------------------------------------------------------
//void swapLayers(std::string Layer1Name, std::string Layer2Name)
//Exchanges the z-order of two layers

void CL_SceneManager::swapLayers(std::string Layer1Name, std::string Layer2Name)
{
	unsigned int Layer1 = 0;
	unsigned int Layer2 = 0;

	for(unsigned int i = 0; i < m_Layers.size(); i++)
	{
		if(Layer1Name == m_Layers[i]->getName())
			Layer1 = i;

		if(Layer2Name == m_Layers[i]->getName())
			Layer2 = i;
	}

	swapLayers(Layer1, Layer2);
}

//---------------------------------------------------------------------------
//CL_SceneLayer* removeLayer(unsigned int Index)
//Removes a layer from the scene, but does not delete the actual layer object

CL_SceneLayer* CL_SceneManager::removeLayer(unsigned int Index)
{
	CL_SceneLayer* Layer = m_Layers[Index];
	std::vector<CL_SceneLayer*>::iterator it = m_Layers.begin();
	it += Index;
	m_Layers.erase(it);
	return Layer;
}

//---------------------------------------------------------------------------
//CL_SceneLayer* removeLayer(std::string Name)
//Removes a layer from the scene, but does not delete the actual layer object

CL_SceneLayer* CL_SceneManager::removeLayer(std::string Name)
{
	for(unsigned int i = 0; i < m_Layers.size(); i++)
	{
		if(Name == m_Layers[i]->getName())
		{
			CL_SceneLayer* Layer = m_Layers[i];
			removeLayer(i);
			return Layer;
		}
	}

	return NULL;
}

//---------------------------------------------------------------------------
//CL_SceneObject* findObject(std::string Name)
//Finds first object in the scene with matching name

CL_SceneObject* CL_SceneManager::findObject(std::string Name)
{
	for(unsigned int i = 0; i < m_Layers.size(); i++)
	{
		CL_SceneLayer *Layer = m_Layers[i];

		std::vector<CL_SceneObject*> List = (*Layer->getObjects());

		for(unsigned int j = 0; j < List.size(); j++)
		{
			if(Name == (List[j])->getName())
				return (List[j]);
		}
	}

	return NULL;
}

//---------------------------------------------------------------------------
//void swapObject(CL_SceneLayer* Layer, unsigned int Index1, unsigned int Index2)
//Swaps the z-order of two sprites in a layer

void CL_SceneManager::swapObject(CL_SceneLayer* Layer, unsigned int Index1, unsigned int Index2)
{
	std::vector<CL_SceneObject*> List = (*Layer->getObjects());

	std::swap(List[Index1], List[Index2]);
}

//---------------------------------------------------------------------------
//void swapObject(CL_SceneLayer* Layer, std::string Object1Name, std::string Object2Name)
//Swaps the z-order of two sprites in a layer

void CL_SceneManager::swapObject(CL_SceneLayer* Layer, std::string Object1Name, std::string Object2Name)
{
	std::vector<CL_SceneObject*> List = (*Layer->getObjects());

	unsigned int Object1 = 0;
	unsigned int Object2 = 0;

	for(unsigned int i = 0; i < List.size(); i++)
	{
		if(Object1Name == (List[i])->getName())
			Object1 = i;

		if(Object2Name == (List[i])->getName())
			Object2 = i;
	}

	swapObject(Layer, Object1, Object2);
}

//---------------------------------------------------------------------------
//void swapObject(std::string LayerName, unsigned int Index1, unsigned int Index2)
//Swaps the z-order of two sprites in a layer

void CL_SceneManager::swapObject(std::string LayerName, unsigned int Index1, unsigned int Index2)
{
	CL_SceneLayer* Layer = findLayer(LayerName);

	if(Layer)
		swapObject(Layer, Index1, Index2);
}

//---------------------------------------------------------------------------
//void swapObject(std::string LayerName, std::string Object1Name, std::string Object2Name)
//Swaps the z-order of two sprites in a layer

void CL_SceneManager::swapObject(std::string LayerName, std::string Object1Name, std::string Object2Name)
{
	CL_SceneLayer* Layer = findLayer(LayerName);

	if(Layer)
		swapObject(Layer, Object1Name, Object2Name);
}

//---------------------------------------------------------------------------
//void saveToXMLFile(std::string FileName)
//Saves the scene to an XML file
void CL_SceneManager::saveToXMLFile(std::string FileName)
{
	CL_DomDocument document; //Create XML Document

	CL_DomElement rootelement = document.create_element("scene"); //Root Scene Element
	rootelement.set_attribute("name", m_SceneName);

	CL_DomElement layerselement = document.create_element("layers"); //Root Layers Element

	rootelement.append_child(layerselement); //Append to scene

	std::list<CL_SceneLayer*>::const_iterator it;

	for(unsigned int i = 0; i < m_Layers.size(); i++) //Cycle through layers
	{
		CL_SceneLayer *Layer = m_Layers[i];
		CL_DomElement layerelement = document.create_element("layer");

		layerelement.set_attribute("name", Layer->getName());
		layerelement.set_attribute("posx", ToString(Layer->getX()));
		layerelement.set_attribute("posy", ToString(Layer->getY()));
		layerelement.set_attribute("width", ToString(Layer->get_width()));
		layerelement.set_attribute("height", ToString(Layer->get_height()));

		if(Layer->getVisible())
			layerelement.set_attribute("visible", "true");
		else
			layerelement.set_attribute("visible", "false");

		//cycle layer objects

		std::vector<CL_SceneObject*> List = (*Layer->getObjects());

		for(unsigned int j = 0; j < List.size(); j++)
		{
			CL_SceneObject *Object = List[j];
			CL_DomElement objectelement = document.create_element("sprite");

			objectelement.set_attribute("name", Object->getName());
			objectelement.set_attribute("posx", ToString(Object->getX()));
			objectelement.set_attribute("posy", ToString(Object->getY()));

			if(Object->getVisible())
				objectelement.set_attribute("visible", "true");
			else
				objectelement.set_attribute("visible", "false");

			layerelement.append_child(objectelement);
		}

		layerselement.append_child(layerelement);
	}

	document.append_child(rootelement);
	document.save(new CL_OutputSource_File(FileName), true, true);
}

//---------------------------------------------------------------------------
//void loadFromXMLFile(std::string FileName)
//Loads the scene from an XML file
//Requires a valid m_ResourceManager if sprites are to be loaded

void CL_SceneManager::loadFromXMLFile(std::string FileName)
{
	CL_DomDocument document;
	document.load(new CL_InputSource_File(FileName), true, true);

	CL_DomElement element = document.get_document_element(); //Get Root Element

	m_SceneName = element.get_attribute("name", "");

	CL_DomNodeList LayerRootList = element.get_elements_by_tag_name("layers");

	if(LayerRootList.get_length() > 0)
	{
		CL_DomNode LayerRootNode = LayerRootList.item(0);

		CL_DomNodeList LayerList = LayerRootNode.to_element().get_elements_by_tag_name("layer");

		if(LayerList.get_length() > 0) //if there is one layer or more
		{
			for(int i=0; i < LayerList.get_length(); i++)
			{
				CL_DomElement LayerElement = LayerList.item(i).to_element();

				std::string Name = LayerElement.get_attribute("name", "");
				float PosX, PosY, Width, Height;

				bool LayerVisible = true;

				if(LayerElement.get_attribute("visible", "true") != "true")
					LayerVisible = false;

				FromString(PosX, LayerElement.get_attribute("posx", "0"));
				FromString(PosY, LayerElement.get_attribute("posy", "0"));
				FromString(Width, LayerElement.get_attribute("width", "0"));
				FromString(Height, LayerElement.get_attribute("height", "0"));

				CL_SceneLayer *Layer = createLayer(Width, Height, CL_PixelFormat::rgba8888, Width*4, Name);

				if(m_ResourceManager) //If resource manager is valid to load layer sprites
				{
					CL_DomNodeList ObjectList = LayerElement.get_elements_by_tag_name("sprite"); //Get Sprite List

					if(ObjectList.get_length() > 0)
					{
						for(int j=0; j < ObjectList.get_length(); j++)
						{
							CL_DomElement ObjectElement = ObjectList.item(j).to_element();

							std::string SpriteName = ObjectElement.get_attribute("name", "");
							float ObjectPosX, ObjectPosY;
							bool ObjectVisible = true;

							if(ObjectElement.get_attribute("visible", "true") != "true")
								ObjectVisible = false;

							FromString(ObjectPosX, ObjectElement.get_attribute("posx", "0"));
							FromString(ObjectPosY, ObjectElement.get_attribute("posy", "0"));


							CL_SceneObject* Object = createSceneObject(SpriteName, m_ResourceManager,
													ObjectPosX, ObjectPosY, ObjectVisible);

							Layer->attachObject(Object);
						}
					}
				}

				addLayer(Layer);
			}
		}
	}
}
